package me.km.overrides;

import com.google.common.collect.Lists;
import com.mojang.authlib.GameProfile;
import io.netty.buffer.Unpooled;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import me.km.KajetansMod;
import me.km.networking.ModPacketHandler;
import me.km.utils.ReflectionUtils;
import net.minecraft.entity.Entity;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.entity.player.EntityPlayerMP;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.network.NetHandlerPlayServer;
import net.minecraft.network.NetworkManager;
import net.minecraft.network.PacketBuffer;
import net.minecraft.network.play.server.SPacketChangeGameState;
import net.minecraft.network.play.server.SPacketCustomPayload;
import net.minecraft.network.play.server.SPacketEntityEffect;
import net.minecraft.network.play.server.SPacketHeldItemChange;
import net.minecraft.network.play.server.SPacketJoinGame;
import net.minecraft.network.play.server.SPacketPlayerAbilities;
import net.minecraft.network.play.server.SPacketRespawn;
import net.minecraft.network.play.server.SPacketServerDifficulty;
import net.minecraft.network.play.server.SPacketSetExperience;
import net.minecraft.network.play.server.SPacketSpawnPosition;
import net.minecraft.network.play.server.SPacketTagsList;
import net.minecraft.network.play.server.SPacketUpdateRecipes;
import net.minecraft.potion.PotionEffect;
import net.minecraft.server.dedicated.DedicatedPlayerList;
import net.minecraft.server.dedicated.DedicatedServer;
import net.minecraft.server.management.DemoPlayerInteractionManager;
import net.minecraft.server.management.PlayerInteractionManager;
import net.minecraft.server.management.PlayerList;
import net.minecraft.server.management.PlayerProfileCache;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.text.TextComponentTranslation;
import net.minecraft.world.IWorld;
import net.minecraft.world.WorldServer;
import net.minecraft.world.chunk.storage.AnvilChunkLoader;
import net.minecraft.world.dimension.DimensionType;
import net.minecraft.world.storage.WorldInfo;
import org.apache.logging.log4j.LogManager;

public class ModPlayerList extends DedicatedPlayerList
{
    private final DedicatedServer server;
    private final List<EntityPlayerMP> players;
    private final Map<UUID, EntityPlayerMP> uuidToPlayerMap;
    
    public ModPlayerList(DedicatedServer server)
    {
        super(server);
        this.server = server;
        this.players = getPlayers();
        this.uuidToPlayerMap = ReflectionUtils.getFieldValue(Map.class, PlayerList.class, this, "field_177454_f");
    }
    
    @Override
    public void initializeConnectionToPlayer(NetworkManager netManager, EntityPlayerMP p)
    {
        GameProfile profile = p.getGameProfile();
        PlayerProfileCache cache = this.server.getPlayerProfileCache();
        GameProfile cachedProfile = cache.getProfileByUUID(profile.getId());
        String s = cachedProfile == null ? profile.getName() : cachedProfile.getName();
        cache.addEntry(profile);
        NBTTagCompound tag = this.readPlayerDataFromFile(p);

        //Forge: Make sure the dimension hasn't been deleted, if so stick them in the overworld.
        DimensionType type = p.dimension;
        boolean noSchedule = true;
        if(type == null || !type.isVanilla())
        {
            WorldServer overWorld = this.server.getWorld(DimensionType.OVERWORLD);
            p.dimension = DimensionType.OVERWORLD;
            if(type != null && this.server.getWorld(type) != null)
            {
                double x = p.posX;
                double y = p.posY;
                double z = p.posZ;
                float yaw = p.rotationYaw;
                float pitch = p.rotationPitch;
                noSchedule = false;
                KajetansMod.scheduler.scheduleTask(() -> 
                {
                    WorldServer ws = server.getWorld(type);
                    if(ws != null)
                    {
                        p.teleport(ws, x, y, z, yaw, pitch);
                    }       
                    net.minecraftforge.fml.hooks.BasicEventHooks.firePlayerLoggedIn(p);
                }, 2);
            }
            WorldInfo wi = overWorld.getWorldInfo();
            p.setPosition(wi.getSpawnX(), wi.getSpawnY(), wi.getSpawnZ());
        }
        
        WorldServer ws = this.server.getWorld(p.dimension);
        p.setWorld(ws);
        p.interactionManager.setWorld((WorldServer) p.world);
        String s1 = "local";
        if(netManager.getRemoteAddress() != null)
        {
            s1 = netManager.getRemoteAddress().toString();
        }

        LogManager.getLogger().info("{}[{}] logged in with entity id {} at ({}, {}, {})", p.getName().getString(), s1, p.getEntityId(), p.posX, p.posY, p.posZ);
        WorldInfo wi = ws.getWorldInfo();
        this.setPlayerGameTypeBasedOnOther(p, null, ws);
        NetHandlerPlayServer nhps = new NetHandlerPlayServer(this.server, netManager, p);
        ModPacketHandler.syncDimensions(p);
        nhps.sendPacket(new SPacketJoinGame(p.getEntityId(), p.interactionManager.getGameType(), wi.isHardcore(), ws.dimension.getType(), ws.getDifficulty(), this.getMaxPlayers(), wi.getTerrainType(), ws.getGameRules().getBoolean("reducedDebugInfo")));
        nhps.sendPacket(new SPacketCustomPayload(SPacketCustomPayload.BRAND, (new PacketBuffer(Unpooled.buffer())).writeString(this.getServer().getServerModName())));
        nhps.sendPacket(new SPacketServerDifficulty(wi.getDifficulty(), wi.isDifficultyLocked()));
        nhps.sendPacket(new SPacketPlayerAbilities(p.abilities));
        nhps.sendPacket(new SPacketHeldItemChange(p.inventory.currentItem));
        nhps.sendPacket(new SPacketUpdateRecipes(this.server.getRecipeManager().getRecipes()));
        nhps.sendPacket(new SPacketTagsList(this.server.getNetworkTagManager()));
        this.updatePermissionLevel(p);
        p.getStats().markAllDirty();
        p.getRecipeBook().init(p);
        this.sendScoreboard(ws.getScoreboard(), p);
        this.server.refreshStatusNextTick();

        this.playerLoggedIn(p);
        nhps.setPlayerLocation(p.posX, p.posY, p.posZ, p.rotationYaw, p.rotationPitch);
        this.sendWorldInfo(p, ws);
        if(!this.server.getResourcePackUrl().isEmpty())
        {
            p.loadResourcePack(this.server.getResourcePackUrl(), this.server.getResourcePackHash());
        }

        for(PotionEffect potioneffect : p.getActivePotionEffects())
        {
            nhps.sendPacket(new SPacketEntityEffect(p.getEntityId(), potioneffect));
        }

        if(tag != null && tag.contains("RootVehicle", 10))
        {
            NBTTagCompound nbttagcompound1 = tag.getCompound("RootVehicle");
            Entity entity1 = AnvilChunkLoader.readWorldEntity(nbttagcompound1.getCompound("Entity"), ws, true);
            if(entity1 != null)
            {
                UUID uuid = nbttagcompound1.getUniqueId("Attach");
                if(entity1.getUniqueID().equals(uuid))
                {
                    p.startRiding(entity1, true);
                }
                else
                {
                    for(Entity entity : entity1.getRecursivePassengers())
                    {
                        if(entity.getUniqueID().equals(uuid))
                        {
                            p.startRiding(entity, true);
                            break;
                        }
                    }
                }

                if(!p.isPassenger())
                {
                    LogManager.getLogger().warn("Couldn't reattach entity to player");
                    ws.removeEntityDangerously(entity1);

                    for(Entity entity2 : entity1.getRecursivePassengers())
                    {
                        ws.removeEntityDangerously(entity2);
                    }
                }
            }
        }

        p.addSelfToInternalCraftingInventory();
        if(noSchedule)
        {
            net.minecraftforge.fml.hooks.BasicEventHooks.firePlayerLoggedIn(p);
        }
    }

    @Override
    public EntityPlayerMP createPlayerForUser(GameProfile profile)
    {
        UUID uuid = EntityPlayer.getUUID(profile);
        List<EntityPlayerMP> list = Lists.newArrayList();

        for(int i = 0; i < this.players.size(); ++i)
        {
            EntityPlayerMP entityplayermp = this.players.get(i);
            if(entityplayermp.getUniqueID().equals(uuid))
            {
                list.add(entityplayermp);
            }
        }

        EntityPlayerMP entityplayermp2 = this.uuidToPlayerMap.get(profile.getId());
        if(entityplayermp2 != null && !list.contains(entityplayermp2))
        {
            list.add(entityplayermp2);
        }

        for(EntityPlayerMP entityplayermp1 : list)
        {
            entityplayermp1.connection.disconnect(new TextComponentTranslation("multiplayer.disconnect.duplicate_login"));
        }

        PlayerInteractionManager playerinteractionmanager;
        if(this.server.isDemo())
        {
            playerinteractionmanager = new DemoPlayerInteractionManager(this.server.getWorld(DimensionType.OVERWORLD));
        }
        else
        {
            playerinteractionmanager = new PlayerInteractionManager(this.server.getWorld(DimensionType.OVERWORLD));
        }

        return new ModEntityPlayerMP(this.server, this.server.getWorld(DimensionType.OVERWORLD), profile, playerinteractionmanager);
    }

    @Override
    public EntityPlayerMP recreatePlayerEntity(EntityPlayerMP pIn, DimensionType dim, boolean conqueredEnd)
    {
        WorldServer world = server.getWorld(dim);
        if(world == null)
        {
            dim = pIn.getSpawnDimension();
        }
        else if(!world.getDimension().canRespawnHere())
        {
            dim = world.getDimension().getRespawnDimension(pIn);
        }
        if(server.getWorld(dim) == null)
        {
            dim = DimensionType.OVERWORLD;
        }

        pIn.getServerWorld().getEntityTracker().removePlayerFromTrackers(pIn);
        pIn.getServerWorld().getEntityTracker().untrack(pIn);
        pIn.getServerWorld().getPlayerChunkMap().removePlayer(pIn);
        this.players.remove(pIn);
        this.server.getWorld(pIn.dimension).removeEntityDangerously(pIn, true); // Forge: keep data until copyFrom called
        BlockPos blockpos = pIn.getBedLocation(dim);
        boolean flag = pIn.isSpawnForced(dim);
        pIn.dimension = dim;
        PlayerInteractionManager playerinteractionmanager;
        if(this.server.isDemo())
        {
            playerinteractionmanager = new DemoPlayerInteractionManager(this.server.getWorld(pIn.dimension));
        }
        else
        {
            playerinteractionmanager = new PlayerInteractionManager(this.server.getWorld(pIn.dimension));
        }

        EntityPlayerMP p = new ModEntityPlayerMP(this.server, this.server.getWorld(pIn.dimension), pIn.getGameProfile(), playerinteractionmanager);
        p.connection = pIn.connection;
        p.copyFrom(pIn, conqueredEnd);
        pIn.remove(false); // Forge: clone event had a chance to see old data, now discard it
        p.dimension = dim;
        p.setEntityId(pIn.getEntityId());
        p.setPrimaryHand(pIn.getPrimaryHand());

        for(String s : pIn.getTags())
        {
            p.addTag(s);
        }

        WorldServer worldserver = this.server.getWorld(pIn.dimension);
        this.setPlayerGameTypeBasedOnOther(p, pIn, worldserver);
        if(blockpos != null)
        {
            BlockPos blockpos1 = EntityPlayer.getBedSpawnLocation(this.server.getWorld(pIn.dimension), blockpos, flag);
            if(blockpos1 != null)
            {
                p.setLocationAndAngles((double) ((float) blockpos1.getX() + 0.5F), (double) ((float) blockpos1.getY() + 0.1F), (double) ((float) blockpos1.getZ() + 0.5F), 0.0F, 0.0F);
                p.setSpawnPoint(blockpos, flag, pIn.dimension);
            }
            else
            {
                p.connection.sendPacket(new SPacketChangeGameState(0, 0.0F));
            }
        }

        worldserver.getChunkProvider().provideChunk((int) p.posX >> 4, (int) p.posZ >> 4, true, true);

        while(!worldserver.isCollisionBoxesEmpty(p, p.getBoundingBox()) && p.posY < 256.0D)
        {
            p.setPosition(p.posX, p.posY + 1.0D, p.posZ);
        }

        p.connection.sendPacket(new SPacketRespawn(p.dimension, world.getDifficulty(), world.getWorldInfo().getTerrainType(), p.interactionManager.getGameType())); // Forge: Use new dimensions information 
        BlockPos blockpos2 = worldserver.getSpawnPoint();
        p.connection.setPlayerLocation(p.posX, p.posY, p.posZ, p.rotationYaw, p.rotationPitch);
        p.connection.sendPacket(new SPacketSpawnPosition(blockpos2));
        p.connection.sendPacket(new SPacketSetExperience(p.experience, p.experienceTotal, p.experienceLevel));
        this.sendWorldInfo(p, worldserver);
        this.updatePermissionLevel(p);
        worldserver.getPlayerChunkMap().addPlayer(p);
        worldserver.spawnEntity(p);
        this.players.add(p);
        this.uuidToPlayerMap.put(p.getUniqueID(), p);
        p.addSelfToInternalCraftingInventory();
        p.setHealth(p.getHealth());
        net.minecraftforge.fml.hooks.BasicEventHooks.firePlayerRespawnEvent(p, conqueredEnd);
        return p;
    }
    
    private void setPlayerGameTypeBasedOnOther(EntityPlayerMP target, EntityPlayerMP source, IWorld w) 
    {
        ReflectionUtils.setPlayerGameTypeBasedOnOther(this, target, source, w);
    }
}
